/*
 * Common clock functionality.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>

#include "cpm-common.h"
#include <math.h>
#include <stdio.h>



// Register access.

Cpm_regs::Cpm_regs(l4_addr_t addr, Clock_base *clocks[])
: _clocks(clocks)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
}

// Utility methods.

uint32_t
Cpm_regs::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_regs::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

Clock_base *
Cpm_regs::get_clock(int num)
{
  return _clocks[num];
}



// Field methods.

uint32_t
Field::get_field(Cpm_regs &regs)
{
  if (defined)
    return regs.get_field(reg, mask, bit);
  else
    return 0;
}

void
Field::set_field(Cpm_regs &regs, uint32_t value)
{
  if (defined)
    regs.set_field(reg, mask, bit, value);
}

// Undefined field.

Field Field::undefined;



// Clock sources.

enum Clock_identifiers
Mux::get_input(int num)
{
  if (num < _num_inputs)
    return _inputs[num];
  else
    return Clock_undefined;
}

// Clock sources.

uint8_t
Source::get_source(Cpm_regs &regs)
{
  if (_source.is_defined())
    return _source.get_field(regs);
  else
    return 0;
}

void
Source::set_source(Cpm_regs &regs, uint8_t source)
{
  if (!_source.is_defined())
    return;

  _source.set_field(regs, source);
}

enum Clock_identifiers
Source::get_source_clock(Cpm_regs &regs)
{
  return get_input(get_number() == 1 ? 0 : get_source(regs));
}

void
Source::set_source_clock(Cpm_regs &regs, enum Clock_identifiers clock)
{
  for (int source = 0; source < _inputs.get_number(); source++)
    if (get_input(source) == clock)
      _source.set_field(regs, source);
}

// Clock source frequencies.

uint64_t
Source::get_frequency(Cpm_regs &regs)
{
  enum Clock_identifiers input = get_source_clock(regs);

  if (input != Clock_undefined)
    return regs.get_clock(input)->get_frequency(regs);
  else
    return 0;
}



// Clock control.

Control_base::~Control_base()
{
}

void
Control_base::change_disable(Cpm_regs &regs)
{
  (void) regs;
}

void
Control_base::change_enable(Cpm_regs &regs)
{
  (void) regs;
}

int
Control::have_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    return _gate.get_field(regs) == _gate.get_asserted();
  else
    return true;
}

void
Control::start_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, _gate.get_asserted());
}

void
Control::stop_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, _gate.get_deasserted());
}

void
Control::wait_busy(Cpm_regs &regs)
{
  if (_busy.is_defined())
    while (_busy.get_field(regs));
}

void
Control::change_disable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 0);
}

void
Control::change_enable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 1);
}

// Undefined control.

Control Control::undefined;



// PLL-specific control.

int
Control_pll::have_pll(Cpm_regs &regs)
{
  return _stable.get_field(regs);
}

int
Control_pll::pll_enabled(Cpm_regs &regs)
{
  return _enable.get_field(regs);
}

int
Control_pll::pll_bypassed(Cpm_regs &regs)
{
  return _bypass.get_field(regs);
}

void
Control_pll::pll_bypass(Cpm_regs &regs)
{
  _bypass.set_field(regs, 1);
}

void
Control_pll::pll_engage(Cpm_regs &regs)
{
  _bypass.set_field(regs, 0);
}

// Clock control.

int
Control_pll::have_clock(Cpm_regs &regs)
{
  return have_pll(regs) && pll_enabled(regs);
}

void
Control_pll::start_clock(Cpm_regs &regs)
{
  _enable.set_field(regs, 1);
  while (!have_pll(regs));
}

void
Control_pll::stop_clock(Cpm_regs &regs)
{
  _enable.set_field(regs, 0);
  while (have_pll(regs));
}

void
Control_pll::wait_busy(Cpm_regs &regs)
{
  // NOTE: Could wait for some kind of stable or "lock" signal, but the chips
  //       provide all sorts of differing signals.

  (void) regs;
}

void
Control_pll::change_disable(Cpm_regs &regs)
{
  if (_enabled)
    start_clock(regs);
}

void
Control_pll::change_enable(Cpm_regs &regs)
{
  // NOTE: Since the X1600 manual warns of changing the frequency while enabled,
  //       it is easier to just stop and then start the PLL again.

  _enabled = have_clock(regs);

  if (_enabled)
    stop_clock(regs);
}



// Clock dividers.

Divider_base::~Divider_base()
{
}



// Fixed divider.

uint64_t
Divider_fixed::get_frequency(Cpm_regs &regs, uint64_t source_frequency)
{
  (void) regs;
  return source_frequency / _value;
}

int
Divider_fixed::set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency)
{
  (void) regs; (void) source_frequency; (void) frequency;
  return 0;
}

int
Divider_fixed::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  (void) regs;
  parameters[0] = _value;
  return 1;
}

int
Divider_fixed::set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[])
{
  (void) regs; (void) num_parameters; (void) parameters;
  return 0;
}



// Simple divider for regular clocks.

uint32_t
Divider::get_divider(Cpm_regs &regs)
{
  if (_divider.is_defined())
    return _divider.get_field(regs) + 1;
  else
    return 1;
}

void
Divider::set_divider(Cpm_regs &regs, uint32_t divider)
{
  if (_divider.is_defined())
    _divider.set_field(regs, divider - 1);
}

// Output clock frequencies.

uint64_t
Divider::get_frequency(Cpm_regs &regs, uint64_t source_frequency)
{
  return source_frequency / get_divider(regs);
}

int
Divider::set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency)
{
  set_divider(regs, (uint32_t) round((double) source_frequency / (double) frequency));
  return 1;
}

int
Divider::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  parameters[0] = get_divider(regs);
  return 1;
}

int
Divider::set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[])
{
  if (num_parameters)
  {
    set_divider(regs, parameters[0]);
    return 1;
  }

  return 0;
}



// Common divider functionality.

static int is_integer(double x)
{
  double target = round(x) * 1000;
  double rounded = floor(x * 1000);

  return (target - 100 < rounded) && (rounded < target + 100);
}

static double getscale_part(double x)
{
  double part = x - floor(x);

  if (part > 0.5)
    return 1 / (1 - part);
  else if (part > 0)
    return 1 / part;
  else
    return 1;
}

static double getscale(double x)
{
  double scale = getscale_part(x);

  if (is_integer(scale))
    return scale;
  else
    return scale * getscale(scale);
}

static void get_divider_operands(double frequency, double source_frequency,
                                 double *multiplier, double *divider)
{
  double ratio = frequency / source_frequency;
  double scale = getscale(ratio);

  *multiplier = scale * ratio;
  *divider = scale;
}

static void reduce_divider_operands(uint32_t *m, uint32_t *n, uint32_t m_limit,
                                    uint32_t n_limit)
{
  while ((*m > m_limit) && (*n > n_limit) && (*m > 1) && (*n > 1))
  {
    *m >>= 1;
    *n >>= 1;
  }
}



#define zero_as_one(X) ((X) ? (X) : 1)

// Feedback multiplier.

uint32_t
Divider_pll::get_multiplier(Cpm_regs &regs)
{
  if (_adjust_by_one)
    return _multiplier.get_field(regs) + 1;
  else
    return zero_as_one(_multiplier.get_field(regs));
}

void
Divider_pll::set_multiplier(Cpm_regs &regs, uint32_t multiplier)
{
  if (_adjust_by_one)
    _multiplier.set_field(regs, multiplier - 1);
  else
    _multiplier.set_field(regs, multiplier);
}

// Input divider.

uint32_t
Divider_pll::get_input_divider(Cpm_regs &regs)
{
  if (_adjust_by_one)
    return _input_divider.get_field(regs) + 1;
  else
    return zero_as_one(_input_divider.get_field(regs));
}

void
Divider_pll::set_input_divider(Cpm_regs &regs, uint32_t divider)
{
  if (_adjust_by_one)
    _input_divider.set_field(regs, divider - 1);
  else
    _input_divider.set_field(regs, divider);
}

// Output dividers.

uint32_t
Divider_pll::get_output_divider(Cpm_regs &regs)
{
  uint32_t d0, d1;

  if (_adjust_by_one)
  {
    d0 = _output_divider0.get_field(regs) + 1;
    d1 = _output_divider1.is_defined() ?
         _output_divider1.get_field(regs) + 1 : 1;
  }
  else
  {
    d0 = zero_as_one(_output_divider0.get_field(regs));
    d1 = _output_divider1.is_defined() ?
         zero_as_one(_output_divider1.get_field(regs)) : 1;
  }

  return d0 * d1;
}

void
Divider_pll::set_output_divider(Cpm_regs &regs, uint32_t divider)
{
  uint32_t d0, d1;

  // Assert 1 as a minimum.

  if (!divider)
    divider = 1;

  // Attempt to set any single divider.

  if (!_output_divider1.is_defined())
  {
    if (_adjust_by_one)
      _output_divider0.set_field(regs, divider);
    else
      _output_divider0.set_field(regs, divider - 1);
    return;
  }

  // For two-divider implementations such as the X1600, divider 0 must be less
  // than or equal to divider 1.

  if (divider < _output_divider1.get_limit())
  {
    d0 = 1;
    d1 = divider;
  }
  else
  {
    d0 = (uint32_t) floor(sqrt(divider));
    d1 = divider / d0;
  }

  if (_adjust_by_one)
  {
    _output_divider0.set_field(regs, d0 - 1);
    _output_divider1.set_field(regs, d1 - 1);
  }
  else
  {
    _output_divider0.set_field(regs, d0);
    _output_divider1.set_field(regs, d1);
  }
}

uint64_t
Divider_pll::get_frequency(Cpm_regs &regs, uint64_t source_frequency)
{
  return (source_frequency * get_multiplier(regs)) /
         (get_input_divider(regs) * get_output_divider(regs));
}

int
Divider_pll::set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency)
{
  double intermediate_multiplier, intermediate_input_divider;
  uint32_t output_min, output_max, output0, output1;
  uint32_t multiplier, input_divider, output_divider;

  // Define the range for the output dividers using the intermediate frequency
  // range applying to each chip, this being the result of the multiplier and
  // input divider.

  output_min = (uint32_t) ceil(_intermediate_min / frequency);
  output_max = (uint32_t) floor(_intermediate_max / frequency);

  // Distribute the divider across the input and output dividers.

  output_divider = output_min;

  while (output_divider <= output_max)
  {
    bool usable_divider;

    // Test divider constraints.

    if (_output_divider1.is_defined())
    {
      output0 = (uint32_t) floor(sqrt(output_divider));
      output1 = (uint32_t) floor(output_divider / output0);

      usable_divider = ((output0 * output1 == output_divider) &&
        (output0 <= _output_divider0.get_limit()) &&
        (output1 <= _output_divider1.get_limit()));
    }
    else
      usable_divider = output_divider <= _output_divider0.get_limit();

    // Apply any usable divider.
    
    if (usable_divider)
    {
      // Calculate the other parameters. Start by working back from the desired
      // output frequency to obtain an intermediate frequency using the proposed
      // divider.

      double intermediate_frequency = frequency * output_divider;

      // Calculate the required multiplier and divider.

      get_divider_operands(intermediate_frequency, source_frequency,
                           &intermediate_multiplier, &intermediate_input_divider);

      multiplier = (uint32_t) round(intermediate_multiplier);
      input_divider = (uint32_t) round(intermediate_input_divider);

      // Attempt to reduce the multiplier and divider to usable values.

      uint32_t multiplier_limit = _multiplier.get_limit();
      uint32_t input_divider_limit = _input_divider.get_limit();

      reduce_divider_operands(&multiplier, &input_divider,
                              multiplier_limit, input_divider_limit);

      if ((multiplier <= multiplier_limit) && (input_divider <= input_divider_limit))
      {
        set_multiplier(regs, multiplier);
        set_input_divider(regs, input_divider);
        set_output_divider(regs, output_divider);

        return 1;
      }
    }

    output_divider++;
  }

  return 0;
}

int
Divider_pll::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  parameters[0] = get_multiplier(regs);
  parameters[1] = get_input_divider(regs);
  parameters[2] = get_output_divider(regs);
  return 3;
}

int
Divider_pll::set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[])
{
  if (num_parameters > 2)
  {
    set_multiplier(regs, parameters[0]);
    set_input_divider(regs, parameters[1]);
    set_output_divider(regs, parameters[2]);

    return 3;
  }

  return 0;
}



// I2S clock divider.

uint32_t
Divider_i2s::get_multiplier(Cpm_regs &regs)
{
  return _multiplier.get_field(regs);
}

uint32_t
Divider_i2s::get_divider_N(Cpm_regs &regs)
{
  return _divider_N.get_field(regs);
}

uint32_t
Divider_i2s::get_divider_D(Cpm_regs &regs)
{
  return _divider_D.get_field(regs);
}

uint64_t
Divider_i2s::get_frequency(Cpm_regs &regs, uint64_t source_frequency)
{
  /* NOTE: Assuming that this is the formula, given that the manual does not
           really describe how D is used. */

  return (source_frequency * get_multiplier(regs)) /
         (get_divider_N(regs) * get_divider_D(regs));
}

int
Divider_i2s::set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency)
{
  double m, n;

  get_divider_operands(frequency, source_frequency, &m, &n);

  uint32_t multiplier = (uint32_t) round(m);
  uint32_t divider = (uint32_t) round(n);

  reduce_divider_operands(&multiplier, &divider,
                          _multiplier.get_limit(),
                          _divider_N.get_limit());

  // Test for operand within limits and the N >= 2M constraint.

  if ((multiplier <= _multiplier.get_limit()) && (divider <= _divider_N.get_limit()) &&
      (divider >= 2 * multiplier))
  {
    /* NOTE: Setting D to 1. Even though it seems that D might also be used,
             it does not seem necessary in practice, and the documentation is
             unclear about its use. */

    uint32_t parameters[] = {multiplier, divider, 1};

    set_parameters(regs, 3, parameters);
    return 1;
  }

  return 0;
}

int
Divider_i2s::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  parameters[0] = get_multiplier(regs);
  parameters[1] = get_divider_N(regs);
  parameters[2] = get_divider_D(regs);
  return 3;
}

int
Divider_i2s::set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[])
{
  if (num_parameters == 1)
  {
    // Set automatic N and D value calculation if only one parameter is given.

    _auto_N.set_field(regs, 0);
    _auto_D.set_field(regs, 0);
    _multiplier.set_field(regs, parameters[0]);

    return 1;
  }
  else if (num_parameters > 1)
  {
    // Require N >= 2M, returning otherwise.

    if (parameters[1] < 2 * parameters[0])
      return 0;

    // Set automatic D value calculation if only two parameters are given.

    _auto_N.set_field(regs, 1);
    _auto_D.set_field(regs, (num_parameters == 2) ? 0 : 1);

    _multiplier.set_field(regs, parameters[0]);
    _divider_N.set_field(regs, parameters[1]);

    // Set D explicitly if given.

    if (num_parameters > 2)
      _divider_D.set_field(regs, parameters[2]);

    return num_parameters;
  }

  return 0;
}



// Clock interface.

Clock_base::~Clock_base()
{
}



// Null clock.

int
Clock_null::have_clock(Cpm_regs &regs)
{
  (void) regs;
  return false;
}

void
Clock_null::start_clock(Cpm_regs &regs)
{
  (void) regs;
}

void
Clock_null::stop_clock(Cpm_regs &regs)
{
  (void) regs;
}

// Output clock frequencies.

uint64_t
Clock_null::get_frequency(Cpm_regs &regs)
{
  (void) regs;
  return 0;
}



// Passive clock.

int
Clock_passive::have_clock(Cpm_regs &regs)
{
  (void) regs;
  return true;
}

void
Clock_passive::start_clock(Cpm_regs &regs)
{
  (void) regs;
}

void
Clock_passive::stop_clock(Cpm_regs &regs)
{
  (void) regs;
}

// Output clock frequencies.

uint64_t
Clock_passive::get_frequency(Cpm_regs &regs)
{
  (void) regs;
  return _frequency;
}



// Clock control.

int
Clock_controlled::have_clock(Cpm_regs &regs)
{
  return _get_control().have_clock(regs);
}

void
Clock_controlled::start_clock(Cpm_regs &regs)
{
  _get_control().start_clock(regs);
}

void
Clock_controlled::stop_clock(Cpm_regs &regs)
{
  _get_control().stop_clock(regs);
}



// Active clock interface.

Clock_active::~Clock_active()
{
}

// Clock sources.

uint8_t
Clock_active::get_source(Cpm_regs &regs)
{
  return _source.get_source(regs);
}

void
Clock_active::set_source(Cpm_regs &regs, uint8_t source)
{
  _get_control().change_enable(regs);
  _source.set_source(regs, source);
  _get_control().wait_busy(regs);
  _get_control().change_disable(regs);
}

enum Clock_identifiers
Clock_active::get_source_clock(Cpm_regs &regs)
{
  return _source.get_source_clock(regs);
}

void
Clock_active::set_source_clock(Cpm_regs &regs, enum Clock_identifiers clock)
{
  _source.set_source_clock(regs, clock);
}

// Clock source frequencies.

uint64_t
Clock_active::get_source_frequency(Cpm_regs &regs)
{
  return _source.get_frequency(regs);
}

// Output clock frequencies.

uint64_t
Clock_active::get_frequency(Cpm_regs &regs)
{
  return get_source_frequency(regs);
}



// Divided clock interface.

Clock_divided_base::~Clock_divided_base()
{
}

// Output clock frequencies.

uint64_t
Clock_divided_base::get_frequency(Cpm_regs &regs)
{
  return _get_divider().get_frequency(regs, get_source_frequency(regs));
}

int
Clock_divided_base::set_frequency(Cpm_regs &regs, uint64_t frequency)
{
  _get_control().change_enable(regs);
  int result = _get_divider().set_frequency(regs, get_source_frequency(regs), frequency);
  _get_control().wait_busy(regs);
  _get_control().change_disable(regs);

  return result;
}

int
Clock_divided_base::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  return _get_divider().get_parameters(regs, parameters);
}

int
Clock_divided_base::set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[])
{
  _get_control().change_enable(regs);
  int n = _get_divider().set_parameters(regs, num_parameters, parameters);
  _get_control().wait_busy(regs);
  _get_control().change_disable(regs);

  return n;
}



// PLL functionality.

Pll::~Pll()
{
}

uint64_t
Pll::get_frequency(Cpm_regs &regs)
{
  if (!_control.pll_bypassed(regs))
    return _divider.get_frequency(regs, get_source_frequency(regs));
  else
    return get_source_frequency(regs);
}

int
Pll::set_frequency(Cpm_regs &regs, uint64_t frequency)
{
  int result = Clock_divided_base::set_frequency(regs, frequency);
  _control.pll_engage(regs);

  return result;
}
